/**
 * @copyright MbientLab License (LICENSE.md)
 * @file model.h
 * @brief 
 */
#pragma once

/**
 * Available MetaWear models
 */
typedef enum {
    MBL_MW_MODEL_NA = -1,
    MBL_MW_MODEL_METAWEAR_R,
    MBL_MW_MODEL_METAWEAR_RG,
    MBL_MW_MODEL_METAWEAR_RPRO,
    MBL_MW_MODEL_METAWEAR_C,
    MBL_MW_MODEL_METAWEAR_CPRO,
    MBL_MW_MODEL_METAENV,
    MBL_MW_MODEL_METADETECT,
    MBL_MW_MODEL_METAHEALTH,
    MBL_MW_MODEL_METATRACKER,
    MBL_MW_MODEL_METAMOTION_R,
    MBL_MW_MODEL_METAMOTION_C
} MblMwModel;
