#pragma once

enum class SettingsRegister : uint8_t {
    DEVICE_NAME = 1,
    AD_INTERVAL,
    TX_POWER,
    START_ADVERTISING = 5,
    SCAN_RESPONSE = 7,
    PARTIAL_SCAN_RESPONSE,
    CONNECTION_PARAMS,
    DISCONNECT_EVENT,
    MAC = 0xb,
    BATTERY_STATE = 0xc,
    WHITELIST_FILTER_MODE = 0x13,
    WHITELIST_ADDRESSES = 0x14
};
